searchNodes=[{"doc":"Wasmex is a fast and secure WebAssembly and WASI runtime for Elixir. It enables lightweight WebAssembly containers to be run in your Elixir backend. This is the main module, providing most of the needed API to run WASM binaries. It uses wasmtime to execute WASM binaries through a Rust NIF. Each WASM module must be compiled from a .wasm or '.wat' file. A compiled WASM module can be instantiated which usually happens in a GenServer . To start the GenServer, start_link/1 is used: iex&gt; bytes = File . read! ( TestHelper . wasm_test_file_path ( ) ) iex&gt; { :ok , instance_pid } = Wasmex . start_link ( %{ bytes : bytes } ) iex&gt; Wasmex . call_function ( instance_pid , &quot;sum&quot; , [ 50 , - 8 ] ) { :ok , [ 42 ] } Memory of a WASM instance can be read/written using Wasmex.Memory : iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , store } = Wasmex . store ( pid ) iex&gt; { :ok , memory } = Wasmex . memory ( pid ) iex&gt; index = 4 iex&gt; Wasmex.Memory . set_byte ( store , memory , index , 42 ) iex&gt; Wasmex.Memory . get_byte ( store , memory , index ) 42 See start_link/1 for starting and configuring a WASM instance and call_function/3 for details about calling WASM functions.","ref":"Wasmex.html","title":"Wasmex","type":"module"},{"doc":"Calls a function with the given name and params on the WASM instance and returns its results. Example iex&gt; wat = &quot;(module ...&gt; (func $helloWorld (result i32) (i32.const 42)) ...&gt; (export \\&quot; hello_world \\&quot; (func $helloWorld)) ...&gt; )&quot; iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : wat } ) iex&gt; Wasmex . call_function ( pid , &quot;hello_world&quot; , [ ] ) { :ok , [ 42 ] } String Handling Strings are common candidates for function parameters and return values. However, they can not be used directly when calling WASM functions, because WASM only knows number data types. Since Strings are just &quot;a bunch of bytes&quot;, we can write these bytes into memory and give our WASM function a pointer to that memory location. Strings as Function Parameters Given we have the following Rust function that returns the first byte of a string input compiled to WASM: #[no_mangle] pub extern &quot;C&quot; fn string_first_byte(bytes: *const u8, length: usize) -&gt; u8 { let slice = unsafe { slice::from_raw_parts(bytes, length) }; match slice.first() { Some(&amp;i) =&gt; i, None =&gt; 0, } } This WASM function can be called from Elixir: iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , store } = Wasmex . store ( pid ) iex&gt; { :ok , memory } = Wasmex . memory ( pid ) iex&gt; index = 42 iex&gt; string = &quot;hello, world&quot; iex&gt; Wasmex.Memory . write_binary ( store , memory , index , string ) iex&gt; Wasmex . call_function ( pid , &quot;string_first_byte&quot; , [ index , String . length ( string ) ] ) { :ok , [ 104 ] } # 104 is the letter &quot;h&quot; in ASCII/UTF-8 encoding Please note that Elixir and Rust assume Strings to be valid UTF-8. Take care when handling other encodings. Strings as Function Return Values Given we have the following Rust function compiled to WASM (again, copied from our test code): #[no_mangle] pub extern &quot;C&quot; fn string() -&gt; *const u8 { b&quot;Hello, World!&quot;.as_ptr() } This function returns a pointer to its memory. This memory location contains the String &quot;Hello, World!&quot;. This is how we would receive this String in Elixir: iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , store } = Wasmex . store ( pid ) iex&gt; { :ok , memory } = Wasmex . memory ( pid ) iex&gt; { :ok , [ pointer ] } = Wasmex . call_function ( pid , &quot;string&quot; , [ ] ) iex&gt; Wasmex.Memory . read_string ( store , memory , pointer , 13 ) &quot;Hello, World!&quot; Specifying a timeout The default timeout for call_function is 5 seconds, or 5000 milliseconds. When calling a long-running function, you can specify a timeout value (in milliseconds) for this call. iex&gt; wat = &quot;(module ...&gt; (func $helloWorld (result i32) (i32.const 42)) ...&gt; (export \\&quot; hello_world \\&quot; (func $helloWorld)) ...&gt; )&quot; iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : wat } ) iex&gt; Wasmex . call_function ( pid , &quot;hello_world&quot; , [ ] , 10_000 ) { :ok , [ 42 ] } In the example above, we specify a timeout of 10 seconds.","ref":"Wasmex.html#call_function/4","title":"Wasmex.call_function/4","type":"function"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Wasmex.html#child_spec/1","title":"Wasmex.child_spec/1","type":"function"},{"doc":"Returns whether a function export with the given name exists in the WASM instance. Examples iex&gt; wat = &quot;(module ...&gt; (func $helloWorld (result i32) (i32.const 42)) ...&gt; (export \\&quot; hello_world \\&quot; (func $helloWorld)) ...&gt; )&quot; iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : wat } ) iex&gt; Wasmex . function_exists ( pid , &quot;hello_world&quot; ) true iex&gt; Wasmex . function_exists ( pid , &quot;something_else&quot; ) false","ref":"Wasmex.html#function_exists/2","title":"Wasmex.function_exists/2","type":"function"},{"doc":"Returns the exported Wasmex.Memory of the given WASM instance. Example iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , % Wasmex.Memory { } } = Wasmex . memory ( pid )","ref":"Wasmex.html#memory/1","title":"Wasmex.memory/1","type":"function"},{"doc":"Returns the Wasmex.Module of the WASM instance. Example iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , % Wasmex.Module { } } = Wasmex . module ( pid )","ref":"Wasmex.html#module/1","title":"Wasmex.module/1","type":"function"},{"doc":"Starts a GenServer which compiles and instantiates a WASM module from the given .wasm or .wat bytes. iex&gt; bytes = File . read! ( TestHelper . wasm_test_file_path ( ) ) iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ bytes : bytes } ) Alternatively, a precompiled Wasmex.Module can be passed with its Wasmex.Store : iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; { :ok , module } = Wasmex.Module . compile ( store , &quot;(module)&quot; ) iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ store : store , module : module } ) Imports WASM imports may be given as an additional option. Imports are a map of namespace-name to namespaces. Each namespace is in turn a map of import-name to import. iex&gt; wat = &quot;(module ...&gt; (import \\&quot; IO \\&quot; \\&quot; inspect \\&quot; (func $log (param i32))) ...&gt; )&quot; iex&gt; io_inspect = fn ( %{ _memory : % Wasmex.Memory { } , _caller : % Wasmex.StoreOrCaller { } } = _context , i ) -&gt; ...&gt; IO . inspect ( i ) ...&gt; end iex&gt; imports = %{ ...&gt; IO : %{ ...&gt; inspect : { :fn , [ :i32 ] , [ ] , io_inspect } , ...&gt; } ...&gt; } iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ bytes : wat , imports : imports } ) In the example above, we import the &quot;IO&quot; namespace. That namespace is a map of imports, in this case the inspect function, which is represented with a tuple of: the import type: :fn (a function), the functions parameter types: [:i32] , the functions return types: [] , and the function to be executed: fn (_context, i) -&gt; IO.inspect(i) end The first param the function receives is always the call context: %{ memory : % Wasmex.Memory { } , caller : % Wasmex.StoreOrCaller { } } = context The caller MUST be used instead of a store in Wasmex API functions. Wasmex might deadlock if the store is used instead of the caller (because running the WASM instance holds a Mutex lock on the store so we cannot use that store again during the execution of an imported function). The caller, however, MUST NOT be used outside of the imported functions scope. All other params are regular parameters as specified by the parameter type list. Valid parameter/return types are: :i32 a 32 bit integer :i64 a 64 bit integer :f32 a 32 bit float :f64 a 64 bit float WASI Optionally, modules can be run with WebAssembly System Interface (WASI) support. WASI functions are provided as native implementations by default but could be overridden with Elixir provided functions. iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ bytes : &quot;(module)&quot; , wasi : true } ) It is possible to overwrite the default WASI functions using the imports map: iex&gt; imports = %{ ...&gt; wasi_snapshot_preview1 : %{ ...&gt; random_get : { :fn , [ :i32 , :i32 ] , [ :i32 ] , ...&gt; fn %{ memory : memory , caller : caller } , address , size -&gt; ...&gt; Enum . each ( 0 .. size , fn index -&gt; ...&gt; Wasmex.Memory . set_byte ( caller , memory , address + index , 0 ) ...&gt; end ) ...&gt; # We chose `4` as the random number with a fair dice roll ...&gt; Wasmex.Memory . set_byte ( caller , memory , address , 4 ) ...&gt; 0 ...&gt; end ...&gt; } ...&gt; } ...&gt; } iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ bytes : &quot;(module)&quot; , imports : imports } ) In the example above, we overwrite the random_get function which is (as all other WASI functions) implemented in Rust. This way our Elixir implementation of random_get is used instead of the default WASI implementation. Oftentimes, WASI programs need additional inputs like environment variables, arguments, or file system access. These can configured by additional Wasmex.Wasi.WasiOptions : iex&gt; wasi_options = % Wasmex.Wasi.WasiOptions { ...&gt; args : [ &quot;hello&quot; , &quot;from elixir&quot; ] , ...&gt; env : %{ ...&gt; &quot;A_NAME_MAPS&quot; =&gt; &quot;to a value&quot; , ...&gt; &quot;THE_TEST_WASI_FILE&quot; =&gt; &quot;prints all environment variables&quot; ...&gt; } , ...&gt; preopen : [ % Wasmex.Wasi.PreopenOptions { path : &quot;lib&quot; , alias : &quot;src&quot; } ] ...&gt; } iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ bytes : &quot;(module)&quot; , wasi : wasi_options } ) It is also possible to capture stdout, stdin, or stderr of a WASI program using pipes: iex&gt; { :ok , stdin } = Wasmex.Pipe . new ( ) iex&gt; { :ok , stdout } = Wasmex.Pipe . new ( ) iex&gt; wasi_options = % Wasmex.Wasi.WasiOptions { ...&gt; args : [ &quot;wasmex&quot; , &quot;echo&quot; ] , ...&gt; stdin : stdin , ...&gt; stdout : stdout ...&gt; } iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasi_test_file_path ( ) ) , wasi : wasi_options } ) iex&gt; Wasmex.Pipe . write ( stdin , &quot;Hey! It compiles! Ship it!&quot; ) iex&gt; Wasmex.Pipe . seek ( stdin , 0 ) iex&gt; { :ok , _ } = Wasmex . call_function ( pid , :_start , [ ] ) iex&gt; Wasmex.Pipe . seek ( stdout , 0 ) iex&gt; Wasmex.Pipe . read ( stdout ) &quot;Hey! It compiles! Ship it! \\n &quot; In the example above, we call a WASI program which echoes a line from stdin back to stdout.","ref":"Wasmex.html#start_link/1","title":"Wasmex.start_link/1","type":"function"},{"doc":"Returns the Wasmex.Store of the WASM instance. Example iex&gt; { :ok , pid } = Wasmex . start_link ( %{ bytes : File . read! ( TestHelper . wasm_test_file_path ( ) ) } ) iex&gt; { :ok , % Wasmex.StoreOrCaller { } } = Wasmex . store ( pid )","ref":"Wasmex.html#store/1","title":"Wasmex.store/1","type":"function"},{"doc":"An Engine which is a global context for compilation and management of WASM modules. Engines store global configuration preferences such as compilation settings, enabled features, etc. You'll likely only need at most one of these for a program. You can create an engine with default configuration settings using EngineConfig::default() . Be sure to consult the documentation of Wasmex.EngineConfig for default settings. Example iex&gt; { :ok , _engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { } )","ref":"Wasmex.Engine.html","title":"Wasmex.Engine","type":"module"},{"doc":"Creates a new Wasmex.Engine with default settings. Example iex&gt; _engine = Wasmex.Engine . default ( )","ref":"Wasmex.Engine.html#default/0","title":"Wasmex.Engine.default/0","type":"function"},{"doc":"Creates a new Wasmex.Engine with the specified options. Example iex&gt; { :ok , _engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { } )","ref":"Wasmex.Engine.html#new/1","title":"Wasmex.Engine.new/1","type":"function"},{"doc":"Ahead-of-time (AOT) compiles a WebAssembly module. The bytes provided must be in one of two formats: A binary-encoded WebAssembly module A text-encoded instance of the WebAssembly text format This method may be used to compile a module for use with a different target host. The output of this method may be used with Wasmex.Module.unsafe_deserialize/2 on hosts compatible with the Wasmex.EngineConfig associated with this Wasmex.Engine . The output of this method is safe to send to another host machine for later execution. As the output is already a compiled module, translation and code generation will be skipped and this will improve the performance of constructing a Wasmex.Module from the output of this method. Example iex&gt; { :ok , engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { } ) iex&gt; bytes = File . read! ( TestHelper . wasm_test_file_path ( ) ) iex&gt; { :ok , _serialized_module } = Wasmex.Engine . precompile_module ( engine , bytes )","ref":"Wasmex.Engine.html#precompile_module/2","title":"Wasmex.Engine.precompile_module/2","type":"function"},{"doc":"","ref":"Wasmex.Engine.html#t:t/0","title":"Wasmex.Engine.t/0","type":"type"},{"doc":"Configures a Wasmex.Engine . Options :consume_fuel - Whether or not to consume fuel when executing WASM instructions. This defaults to false . :cranelift_opt_level - Optimization level for the Cranelift code generator. This defaults to :none . :wasm_backtrace_details - Whether or not backtraces in traps will parse debug info in the WASM file to have filename/line number information. This defaults to false . Example iex&gt; _config = % Wasmex.EngineConfig { } ...&gt; |&gt; Wasmex.EngineConfig . consume_fuel ( true ) ...&gt; |&gt; Wasmex.EngineConfig . cranelift_opt_level ( :speed ) ...&gt; |&gt; Wasmex.EngineConfig . wasm_backtrace_details ( false )","ref":"Wasmex.EngineConfig.html","title":"Wasmex.EngineConfig","type":"module"},{"doc":"Configures whether execution of WebAssembly will &quot;consume fuel&quot; to either halt or yield execution as desired. This can be used to deterministically prevent infinitely-executing WebAssembly code by instrumenting generated code to consume fuel as it executes. When fuel runs out a trap is raised. Note that a Wasmex.Store starts with no fuel, so if you enable this option you'll have to be sure to pour some fuel into Wasmex.Store before executing some code. See Wasmex.StoreOrCaller.add_fuel/2 . Example iex&gt; config = % Wasmex.EngineConfig { } ...&gt; |&gt; Wasmex.EngineConfig . consume_fuel ( true ) iex&gt; config . consume_fuel true","ref":"Wasmex.EngineConfig.html#consume_fuel/2","title":"Wasmex.EngineConfig.consume_fuel/2","type":"function"},{"doc":"Configures the Cranelift code generator optimization level. Allows one of the following values: :none - No optimizations performed, minimizes compilation time by disabling most optimizations. :speed - Generates the fastest possible code, but may take longer. :speed_and_size - Similar to speed , but also performs transformations aimed at reducing code size.","ref":"Wasmex.EngineConfig.html#cranelift_opt_level/2","title":"Wasmex.EngineConfig.cranelift_opt_level/2","type":"function"},{"doc":"Configures whether backtraces in traps will parse debug info in the WASM file to have filename/line number information. When enabled this will causes modules to retain debugging information found in WASM binaries. This debug information will be used when a trap happens to symbolicate each stack frame and attempt to print a filename/line number for each WASM frame in the stack trace. Example iex&gt; config = % Wasmex.EngineConfig { } ...&gt; |&gt; Wasmex.EngineConfig . wasm_backtrace_details ( true ) iex&gt; config . wasm_backtrace_details true","ref":"Wasmex.EngineConfig.html#wasm_backtrace_details/2","title":"Wasmex.EngineConfig.wasm_backtrace_details/2","type":"function"},{"doc":"","ref":"Wasmex.EngineConfig.html#t:t/0","title":"Wasmex.EngineConfig.t/0","type":"type"},{"doc":"Instantiates a WASM module and allows calling exported functions on it. In the majority of cases, you will not need to use this module directly but use the main module Wasmex instead. This module expects to be executed within GenServer context which Wasmex sets up.","ref":"Wasmex.Instance.html","title":"Wasmex.Instance","type":"module"},{"doc":"Calls a function the given name exported by the WASM instance with the given params . The WASM function will be invoked asynchronously in a new OS thread. The calling Process/GenServer will receive a {:returned_function_call, result, from} message once execution finishes. The result either is an {:error, reason} or :ok . call_exported_function/5 assumes to be called within a GenServer context, it expects a from argument as given by GenServer.handle_call/3 . from is returned unchanged to allow the wrapping GenServer to reply to their caller. A BadArg exception may be thrown when given unexpected input data. Function parameters Parameters for WASM functions are automatically casted to WASM values. Note that WebAssembly only knows number datatypes (floats and integers of various sizes). You can pass arbitrary data to WebAssembly by writing that data into an instances Wasmex.Memory . The memory/2 function returns the instances memory. Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; Wasmex.Instance . call_exported_function ( store , instance , &quot;sum&quot; , [ 1 , 2 ] , :from ) :ok iex&gt; receive do ...&gt; { :returned_function_call , { :ok , [ 3 ] } , :from } -&gt; :ok ...&gt; after ...&gt; 1000 -&gt; raise &quot;message_expected&quot; ...&gt; end","ref":"Wasmex.Instance.html#call_exported_function/5","title":"Wasmex.Instance.call_exported_function/5","type":"function"},{"doc":"Whether the WASM instance exports a function with the given name . Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; Wasmex.Instance . function_export_exists ( store , instance , &quot;sum&quot; ) true iex&gt; Wasmex.Instance . function_export_exists ( store , instance , &quot;does_not_exist&quot; ) false","ref":"Wasmex.Instance.html#function_export_exists/3","title":"Wasmex.Instance.function_export_exists/3","type":"function"},{"doc":"Returns the Wasmex.Memory of the WASM instance . Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , % Wasmex.Memory { } } = Wasmex.Instance . memory ( store , instance )","ref":"Wasmex.Instance.html#memory/2","title":"Wasmex.Instance.memory/2","type":"function"},{"doc":"Instantiates a WASM module with the given imports. Returns the instantiated WASM instance. The import parameter is a nested map of WASM namespaces. Each namespace consists of a name and a map of function names to function signatures. Function signatures are a tuple of the form {:fn, arg_types, return_types, callback} . Where arg_types and return_types are lists of :i32 , :i64 , :f32 , :f64 . Each callback function receives a context map as the first argument followed by the arguments specified in its signature. context has the following keys: :memory - The default exported Wasmex.Memory of the WASM instance :caller - The caller of the WASM instance which MUST be used instead of a Wasmex.Store in all Wasmex functions called from within the callback. Failure to do so will result in a deadlock. The caller MUST NOT be used outside of the callback. Example This example instantiates a WASM module with one namespace env having three imported functions imported_sum3 , imported_sumf , and imported_void . The imported function imported_sum3 takes three :i32 (32 bit integer) arguments and returns a :i32 number. Its implementation is defined by the callback function fn _context, a, b, c -&gt; a + b + c end . iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; imports = %{ ...&gt; &quot;env&quot; =&gt; ...&gt; %{ ...&gt; &quot;imported_sum3&quot; =&gt; { :fn , [ :i32 , :i32 , :i32 ] , [ :i32 ] , fn _context , a , b , c -&gt; a + b + c end } , ...&gt; &quot;imported_sumf&quot; =&gt; { :fn , [ :f32 , :f32 ] , [ :f32 ] , fn _context , a , b -&gt; a + b end } , ...&gt; &quot;imported_void&quot; =&gt; { :fn , [ ] , [ ] , fn _context -&gt; nil end } ...&gt; } ...&gt; } iex&gt; { :ok , % Wasmex.Instance { } } = Wasmex.Instance . new ( store , module , imports )","ref":"Wasmex.Instance.html#new/3","title":"Wasmex.Instance.new/3","type":"function"},{"doc":"","ref":"Wasmex.Instance.html#t:t/0","title":"Wasmex.Instance.t/0","type":"type"},{"doc":"Memory is a linear array of bytes to store WASM values. The Memory module provides functions to read and write to this array. Memory is accessible through Wasmex.Instance.memory/2 , Wasmex.Memory.from_instance/2 , or as the caller context of an imported function (see Wasmex.Instance.call_exported_function/5 ). iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Instance . memory ( store , instance ) iex&gt; Wasmex.Memory . set_byte ( store , memory , 0 , 42 ) iex&gt; Wasmex.Memory . get_byte ( store , memory , 0 ) 42 WASM memory is organized in pages of 64kb and may be grown by additional pages.","ref":"Wasmex.Memory.html","title":"Wasmex.Memory","type":"module"},{"doc":"Returns the exported memory resource of the given Wasmex.Instance . Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , % Wasmex.Memory { } } = Wasmex.Memory . from_instance ( store , instance )","ref":"Wasmex.Memory.html#from_instance/2","title":"Wasmex.Memory.from_instance/2","type":"function"},{"doc":"Returns the byte at the given index . Example Set a value at memory position 0 and read it back: iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . set_byte ( store , memory , 0 , 42 ) iex&gt; Wasmex.Memory . get_byte ( store , memory , 0 ) 42","ref":"Wasmex.Memory.html#get_byte/3","title":"Wasmex.Memory.get_byte/3","type":"function"},{"doc":"Grows the amount of available memory by the given number of pages. Returns the number of previously available pages. A page has a size of 64 kB or 65,536 bytes. Returns an error if memory could not be grown. Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . grow ( store , memory , 1 ) 17","ref":"Wasmex.Memory.html#grow/3","title":"Wasmex.Memory.grow/3","type":"function"},{"doc":"Reads the given number of bytes from the given memory at the given index. Returns the read bytes as a binary. Example Reads 5 bytes from memory position 0 , given it contains the 5 ASCII characters forming &quot;hello&quot;. iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . write_binary ( store , memory , 0 , &quot;hello&quot; ) iex&gt; Wasmex.Memory . read_binary ( store , memory , 0 , 5 ) &quot;hello&quot; iex&gt; Wasmex.Memory . read_binary ( store , memory , 3 , 2 ) &quot;lo&quot;","ref":"Wasmex.Memory.html#read_binary/4","title":"Wasmex.Memory.read_binary/4","type":"function"},{"doc":"Reads the given number of bytes from the given memory at the given index. Returns the read bytes as a string. Example Reads 5 bytes from memory position 0 , given it contains the 5 ASCII characters forming &quot;hello&quot;. iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . write_binary ( store , memory , 0 , &quot;hello&quot; ) iex&gt; Wasmex.Memory . read_string ( store , memory , 0 , 5 ) &quot;hello&quot; iex&gt; Wasmex.Memory . read_string ( store , memory , 3 , 2 ) &quot;lo&quot;","ref":"Wasmex.Memory.html#read_string/4","title":"Wasmex.Memory.read_string/4","type":"function"},{"doc":"Sets the byte at the given index to the given value . Example Set a value at memory position 0 and read it back: iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . set_byte ( store , memory , 0 , 42 ) iex&gt; Wasmex.Memory . get_byte ( store , memory , 0 ) 42","ref":"Wasmex.Memory.html#set_byte/4","title":"Wasmex.Memory.set_byte/4","type":"function"},{"doc":"Returns the size in bytes of the given memory. Note that the size of the memory is always a multiple of 64kb (one page). Example iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . size ( store , memory ) 1114112 # in bytes (17 pages of 64 kB)","ref":"Wasmex.Memory.html#size/2","title":"Wasmex.Memory.size/2","type":"function"},{"doc":"Writes the given binary into the memory at the given index . Example Writes 5 bytes representing the ASCII characters for &quot;hello&quot; at memory position 0 . iex&gt; %{ store : store , module : module } = TestHelper . wasm_module ( ) iex&gt; { :ok , instance } = Wasmex.Instance . new ( store , module , %{ } ) iex&gt; { :ok , memory } = Wasmex.Memory . from_instance ( store , instance ) iex&gt; Wasmex.Memory . write_binary ( store , memory , 0 , &quot;hello&quot; ) :ok","ref":"Wasmex.Memory.html#write_binary/4","title":"Wasmex.Memory.write_binary/4","type":"function"},{"doc":"","ref":"Wasmex.Memory.html#t:t/0","title":"Wasmex.Memory.t/0","type":"type"},{"doc":"A compiled WebAssembly module. A WASM Module contains stateless WebAssembly code that has already been compiled and can be instantiated multiple times.","ref":"Wasmex.Module.html","title":"Wasmex.Module","type":"module"},{"doc":"Compiles a WASM module from it's WASM (a .wasm file) or WAT (a .wat file) representation. Compiled modules can be instantiated using Wasmex.start_link/1 or Instance.new/3 . Since module compilation takes time and resources but instantiation is comparatively cheap, it may be a good idea to compile a module once and instantiate it often if you want to run a WASM binary multiple times. Example Read a WASM file and compile it into a WASM module. Use the compiled module to start a running Wasmex.Instance . iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; { :ok , module } = Wasmex.Module . compile ( store , File . read! ( TestHelper . wasm_test_file_path ( ) ) ) iex&gt; { :ok , _pid } = Wasmex . start_link ( %{ store : store , module : module } ) Modules can be compiled from WAT (WebAssembly Text) format: iex&gt; wat = &quot;(module)&quot; # minimal and not very useful iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; { :ok , % Wasmex.Module { } } = Wasmex.Module . compile ( store , wat )","ref":"Wasmex.Module.html#compile/2","title":"Wasmex.Module.compile/2","type":"function"},{"doc":"Lists all exports of a WASM module. Returns a map which has the exports name (string) as key and export info-tuples as values. Info tuples always start with an atom indicating the exports type: :fn (function) :global :table :memory Further parts of the info tuple vary depending on the type. Example List the exported function &quot;hello_world()&quot; of a WASM module: iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; wat = &quot;(module ...&gt; (func $helloWorld (result i32) (i32.const 42)) ...&gt; (export \\&quot; hello_world \\&quot; (func $helloWorld)) ...&gt; )&quot; iex&gt; { :ok , module } = Wasmex.Module . compile ( store , wat ) iex&gt; Wasmex.Module . exports ( module ) %{ &quot;hello_world&quot; =&gt; { :fn , [ ] , [ :i32 ] } , }","ref":"Wasmex.Module.html#exports/1","title":"Wasmex.Module.exports/1","type":"function"},{"doc":"Lists all imports of a WebAssembly module grouped by their module namespace. Returns a map of namespace names to namespaces with each namespace being a map again. A namespace is a map of imports with the import name as key and and info-tuple as value. Info tuples always start with an atom indicating the imports type: :fn (function) :global :table :memory Further parts of the info tuple vary depending on the type. Example Show that the WASM module imports a function &quot;inspect&quot; from the &quot;IO&quot; namespace: iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; wat = &quot;(module ...&gt; (import \\&quot; IO \\&quot; \\&quot; inspect \\&quot; (func $log (param i32))) ...&gt; )&quot; iex&gt; { :ok , module } = Wasmex.Module . compile ( store , wat ) iex&gt; Wasmex.Module . imports ( module ) %{ &quot;IO&quot; =&gt; %{ &quot;inspect&quot; =&gt; { :fn , [ :i32 ] , [ ] } , } }","ref":"Wasmex.Module.html#imports/1","title":"Wasmex.Module.imports/1","type":"function"},{"doc":"Returns the name of the current module if a name is given. This name is normally set in the WASM bytecode by some compilers. Example iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; wat = &quot;(module $hiFromTheDocs)&quot; # minimal and not very useful WASM module iex&gt; { :ok , module } = Wasmex.Module . compile ( store , wat ) iex&gt; Wasmex.Module . name ( module ) &quot;hiFromTheDocs&quot;","ref":"Wasmex.Module.html#name/1","title":"Wasmex.Module.name/1","type":"function"},{"doc":"Serializes a compiled WASM module into a binary. The generated binary can be deserialized back into a module using unsafe_deserialize/1 . It is unsafe do alter the binary in any way. See unsafe_deserialize/1 for safety considerations. Example Serializes a compiled module: iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; { :ok , module } = Wasmex.Module . compile ( store , File . read! ( TestHelper . wasm_test_file_path ( ) ) ) iex&gt; { :ok , serialized } = Wasmex.Module . serialize ( module ) iex&gt; is_binary ( serialized ) true","ref":"Wasmex.Module.html#serialize/1","title":"Wasmex.Module.serialize/1","type":"function"},{"doc":"Deserializes an in-memory compiled module previously created with Wasmex.Module.serialize/1 or Wasmex.Engine::precompile_module/2 . This function will deserialize the binary blobs emitted by Wasmex.Module.serialize/1 and Wasmex.Engine::precompile_module/2 back into an in-memory Wasmex.Module that's ready to be instantiated. Unsafety This function is marked as unsafe because if fed invalid input or used improperly this could lead to memory safety vulnerabilities. This method should not, for example, be exposed to arbitrary user input. The structure of the binary blob read here is only lightly validated internally in wasmtime . This is intended to be an efficient &quot;rehydration&quot; for a Wasmex.Module which has very few runtime checks beyond deserialization. Arbitrary input could, for example, replace valid compiled code with any other valid compiled code, meaning that this can trivially be used to execute arbitrary code otherwise. For these reasons this function is unsafe . This function is only designed to receive the previous input from Wasmex.Module.serialize/1 and Wasmex.Engine::precompile_module/2 . If the exact output of those functions (unmodified) is passed to this function then calls to this function can be considered safe. It is the caller's responsibility to provide the guarantee that only previously-serialized bytes are being passed in here. Note that this function is designed to be safe receiving output from any compiled version of wasmtime itself. This means that it is safe to feed output from older versions of Wasmtime into this function, in addition to newer versions of wasmtime (from the future!). These inputs will deterministically and safely produce an error. This function only successfully accepts inputs from the same version of wasmtime . (this means that if you cache blobs across versions of wasmtime you can be safely guaranteed that future versions of wasmtime will reject old cache entries). It is best to deserialize modules using an Engine with the same configuration as the one used to serialize/precompile it. Example Serializes a compiled module and deserializes it again: iex&gt; { :ok , store } = Wasmex.Store . new ( ) iex&gt; { :ok , module } = Wasmex.Module . compile ( store , File . read! ( TestHelper . wasm_test_file_path ( ) ) ) iex&gt; { :ok , serialized } = Wasmex.Module . serialize ( module ) iex&gt; { :ok , % Wasmex.Module { } } = Wasmex.Module . unsafe_deserialize ( serialized )","ref":"Wasmex.Module.html#unsafe_deserialize/2","title":"Wasmex.Module.unsafe_deserialize/2","type":"function"},{"doc":"","ref":"Wasmex.Module.html#t:t/0","title":"Wasmex.Module.t/0","type":"type"},{"doc":"A Pipe is a memory buffer that can be used in exchange for a WASM file. Pipes have a read and write position which can be set using seek/2 . Example Pipes can be written to and read from: iex&gt; { :ok , pipe } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Pipe . write ( pipe , &quot;hello&quot; ) { :ok , 5 } iex&gt; Wasmex.Pipe . seek ( pipe , 0 ) iex&gt; Wasmex.Pipe . read ( pipe ) &quot;hello&quot; They can be used to capture stdout/stdin/stderr of WASI programs: iex&gt; { :ok , stdin } = Wasmex.Pipe . new ( ) iex&gt; { :ok , stdout } = Wasmex.Pipe . new ( ) iex&gt; { :ok , stderr } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Store . new_wasi ( % Wasmex.Wasi.WasiOptions { ...&gt; stdin : stdin , ...&gt; stdout : stdout , ...&gt; stderr : stderr , ...&gt; } )","ref":"Wasmex.Pipe.html","title":"Wasmex.Pipe","type":"module"},{"doc":"Creates and returns a new Pipe. Example iex&gt; { :ok , % Pipe { } } = Wasmex.Pipe . new ( )","ref":"Wasmex.Pipe.html#new/0","title":"Wasmex.Pipe.new/0","type":"function"},{"doc":"Reads all available bytes from the Pipe and returns them as a binary. This function does not block if there are no bytes available. Reading starts at the current read position, see seek/2 , and forwards the read position to the end of the Pipe. The read bytes are not erased and can be read again after seeking back. Example iex&gt; { :ok , pipe } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Pipe . write ( pipe , &quot;hello&quot; ) iex&gt; Wasmex.Pipe . read ( pipe ) # current position is at EOL, nothing more to read &quot;&quot; iex&gt; Wasmex.Pipe . seek ( pipe , 0 ) iex&gt; Wasmex.Pipe . read ( pipe ) &quot;hello&quot; iex&gt; Wasmex.Pipe . seek ( pipe , 3 ) iex&gt; Wasmex.Pipe . read ( pipe ) &quot;lo&quot; iex&gt; Wasmex.Pipe . read ( pipe ) &quot;&quot;","ref":"Wasmex.Pipe.html#read/1","title":"Wasmex.Pipe.read/1","type":"function"},{"doc":"Sets the read/write position of the Pipe to the given position. The position is given as a number of bytes from the start of the Pipe. Example iex&gt; { :ok , pipe } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Pipe . write ( pipe , &quot;hello&quot; ) iex&gt; Wasmex.Pipe . seek ( pipe , 0 ) :ok iex&gt; Wasmex.Pipe . read ( pipe ) &quot;hello&quot;","ref":"Wasmex.Pipe.html#seek/2","title":"Wasmex.Pipe.seek/2","type":"function"},{"doc":"Returns the current size of the Pipe in bytes. Example iex&gt; { :ok , pipe } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Pipe . size ( pipe ) 0 iex&gt; Wasmex.Pipe . write ( pipe , &quot;hello&quot; ) iex&gt; Wasmex.Pipe . size ( pipe ) 5","ref":"Wasmex.Pipe.html#size/1","title":"Wasmex.Pipe.size/1","type":"function"},{"doc":"Writes the given binary into the pipe. Writing starts at the current write position, see seek/2 , and forwards it. Example iex&gt; { :ok , pipe } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Pipe . write ( pipe , &quot;hello&quot; ) { :ok , 5 } iex&gt; Wasmex.Pipe . seek ( pipe , 0 ) iex&gt; Wasmex.Pipe . read ( pipe ) &quot;hello&quot;","ref":"Wasmex.Pipe.html#write/2","title":"Wasmex.Pipe.write/2","type":"function"},{"doc":"","ref":"Wasmex.Pipe.html#t:t/0","title":"Wasmex.Pipe.t/0","type":"type"},{"doc":"A Store is a collection of WASM instances and host-defined state. All WASM instances and items will be attached to and refer to a Store. For example instances, functions, globals, and tables are all attached to a Store. Instances are created by instantiating a Module within a Store. Many functions of the Wasmex API require a Store in the form of a Wasmex.StoreOrCaller to be passed in. A Store is intended to be a short-lived object in a program. No form of GC is implemented at this time so once an instance is created within a Store it will not be deallocated until the Store itself is garbage collected. This makes Store unsuitable for creating an unbounded number of instances in it because Store will never release this memory. It’s recommended to have a Store correspond roughly to the lifetime of a “main instance”.","ref":"Wasmex.Store.html","title":"Wasmex.Store","type":"module"},{"doc":"Creates a new WASM store. Returns a Wasmex.StoreOrCaller even though we know it’s definitely a Store. This allows Elixir-provided imported functions, which only have a &quot;Caller&quot;, to use the same Wasmex APIs. Optionally, a Wasmex.StoreLimits can be passed in to limit the resources that can be created within the store. Examples iex&gt; { :ok , % Wasmex.StoreOrCaller { } } = Wasmex.Store . new ( ) iex&gt; limits = % Wasmex.StoreLimits { memory_size : 1_000_000 } iex&gt; { :ok , % Wasmex.StoreOrCaller { } } = Wasmex.Store . new ( limits )","ref":"Wasmex.Store.html#new/2","title":"Wasmex.Store.new/2","type":"function"},{"doc":"Creates a new WASM store with WASI support. Returns a Wasmex.StoreOrCaller even though we know it’s definitely a Store. This allows Elixir-provided imported functions, which only have a &quot;Caller&quot;, to use the same Wasmex APIs. See Wasmex.Wasi.WasiOptions for WASI configuration options. Examples iex&gt; { :ok , % Wasmex.StoreOrCaller { } } = Wasmex.Store . new_wasi ( % Wasmex.Wasi.WasiOptions { } ) iex&gt; wasi_opts = % Wasmex.Wasi.WasiOptions { args : [ &quot;arg1&quot; , &quot;arg2&quot; ] } iex&gt; limits = % Wasmex.StoreLimits { memory_size : 1_000_000 } iex&gt; { :ok , % Wasmex.StoreOrCaller { } } = Wasmex.Store . new_wasi ( wasi_opts , limits )","ref":"Wasmex.Store.html#new_wasi/3","title":"Wasmex.Store.new_wasi/3","type":"function"},{"doc":"Configures limits to limit resource creation within a Wasmex.Store . Whenever resources such as linear memory, tables, or instances are allocated the StoreLimits specified here are enforced. Options :memory_size - The maximum number of bytes a linear memory can grow to. Growing a linear memory beyond this limit will fail. By default, linear memory will not be limited. :table_elements - The maximum number of elements in a table. Growing a table beyond this limit will fail. By default, table elements will not be limited. :instances - The maximum number of instances that can be created for a Wasmex.Store . Module instantiation will fail if this limit is exceeded. This value defaults to 10,000. :tables - The maximum number of tables that can be created for a Wasmex.Store . Module instantiation will fail if this limit is exceeded. This value defaults to 10,000. :memories - The maximum number of linear memories that can be created for a Wasmex.Store . Instantiation will fail with an error if this limit is exceeded. This value defaults to 10,000. Example iex&gt; Wasmex.Store . new ( % Wasmex.StoreLimits { ...&gt; memory_size : 1_000_000 , ...&gt; table_elements : 100_000 , ...&gt; instances : 2 , ...&gt; tables : 10 ...&gt; memories : 10 ...&gt; } )","ref":"Wasmex.StoreLimits.html","title":"Wasmex.StoreLimits","type":"module"},{"doc":"","ref":"Wasmex.StoreLimits.html#t:t/0","title":"Wasmex.StoreLimits.t/0","type":"type"},{"doc":"Either a Wasmex.Store or &quot;Caller&quot; for imported functions. A Store is a collection of WASM instances and host-defined state, see Wasmex.Store . A Caller takes the place of a Store in imported function calls. If a Store is needed in Elixir-provided imported functions, always use the provided Caller because using the Store will cause a deadlock (the running WASM instance locks the Stores Mutex). When configured, a StoreOrCaller can consume fuel to halt or yield execution as desired. See Wasmex.EngineConfig.consume_fuel/2 for more information on fuel consumption.","ref":"Wasmex.StoreOrCaller.html","title":"Wasmex.StoreOrCaller","type":"module"},{"doc":"Adds fuel to this Store for WASM to consume while executing. For this method to work, fuel consumption must be enabled via Wasmex.EngineConfig.consume_fuel/2. By default a Wasmex.Store starts with 0 fuel for WASM to execute with (meaning it will immediately trap and halt execution). This function must be called for the store to have some fuel to allow WebAssembly to execute. Most WASM instructions consume 1 unit of fuel. Some instructions, such as nop , drop , block , and loop , consume 0 units, as any execution cost associated with them involves other instructions which do consume fuel. Note that at this time when fuel is entirely consumed it will cause WASM to trap. ## Errors This function will return an error if fuel consumption is not enabled via Wasmex.EngineConfig.consume_fuel/2`. ## Examples iex&gt; {:ok, engine} = Wasmex.Engine.new(%Wasmex.EngineConfig{consume_fuel: true}) iex&gt; {:ok, store} = Wasmex.Store.new(nil, engine) iex&gt; Wasmex.StoreOrCaller.add_fuel(store, 10) :ok","ref":"Wasmex.StoreOrCaller.html#add_fuel/2","title":"Wasmex.StoreOrCaller.add_fuel/2","type":"function"},{"doc":"Synthetically consumes fuel from this Store. For this method to work fuel, consumption must be enabled via Wasmex.EngineConfig.consume_fuel/2 . WebAssembly execution will automatically consume fuel but if so desired the embedder can also consume fuel manually to account for relative costs of host functions, for example. This function will attempt to consume fuel units of fuel from within this store. If the remaining amount of fuel allows this then {:ok, N} is returned where N is the amount of remaining fuel. Otherwise an error is returned and no fuel is consumed. Errors This function will return an error either if fuel consumption is not enabled via Wasmex.EngineConfig.consume_fuel/2 or if fuel exceeds the amount of remaining fuel within this store. Examples iex&gt; { :ok , engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { consume_fuel : true } ) iex&gt; { :ok , store } = Wasmex.Store . new ( nil , engine ) iex&gt; Wasmex.StoreOrCaller . add_fuel ( store , 10 ) iex&gt; Wasmex.StoreOrCaller . fuel_remaining ( store ) { :ok , 10 }","ref":"Wasmex.StoreOrCaller.html#consume_fuel/2","title":"Wasmex.StoreOrCaller.consume_fuel/2","type":"function"},{"doc":"Returns the amount of fuel consumed by this store's execution so far. Note that fuel, if enabled, must be initially added via Wasmex.StoreOrCaller.add_fuel/2 . Errors If fuel consumption is not enabled via Wasmex.EngineConfig.consume_fuel/2 then this function will return an error tuple. Examples iex&gt; { :ok , engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { consume_fuel : true } ) iex&gt; { :ok , store } = Wasmex.Store . new ( nil , engine ) iex&gt; Wasmex.StoreOrCaller . fuel_consumed ( store ) { :ok , 0 } iex&gt; Wasmex.StoreOrCaller . add_fuel ( store , 10 ) iex&gt; { :ok , _fuel } = Wasmex.StoreOrCaller . consume_fuel ( store , 8 ) iex&gt; Wasmex.StoreOrCaller . fuel_consumed ( store ) { :ok , 8 }","ref":"Wasmex.StoreOrCaller.html#fuel_consumed/1","title":"Wasmex.StoreOrCaller.fuel_consumed/1","type":"function"},{"doc":"Returns the amount of fuel available for future execution of this store. Examples iex&gt; { :ok , engine } = Wasmex.Engine . new ( % Wasmex.EngineConfig { consume_fuel : true } ) iex&gt; { :ok , store } = Wasmex.Store . new ( nil , engine ) iex&gt; Wasmex.StoreOrCaller . add_fuel ( store , 10 ) iex&gt; Wasmex.StoreOrCaller . fuel_remaining ( store ) { :ok , 10 }","ref":"Wasmex.StoreOrCaller.html#fuel_remaining/1","title":"Wasmex.StoreOrCaller.fuel_remaining/1","type":"function"},{"doc":"","ref":"Wasmex.StoreOrCaller.html#t:t/0","title":"Wasmex.StoreOrCaller.t/0","type":"type"},{"doc":"Options for preopening a directory. Likely to be extended with read/write permissions once wasmtime supports them. Options :path - The path to the directory to preopen :alias - The alias to use for the directory. The directory will be available at this path in the WASI filesystem. If not specified, the directory will be available at its real path. Example iex&gt; Wasmex.Store . new_wasi ( % Wasmex.Wasi.WasiOptions { ...&gt; preopen : [ ...&gt; % PreopenOptions { path : &quot;/tmp&quot; , alias : &quot;temp&quot; } ...&gt; ] , ...&gt; } )","ref":"Wasmex.Wasi.PreopenOptions.html","title":"Wasmex.Wasi.PreopenOptions","type":"module"},{"doc":"","ref":"Wasmex.Wasi.PreopenOptions.html#t:t/0","title":"Wasmex.Wasi.PreopenOptions.t/0","type":"type"},{"doc":"Configures WASI support for a Wasmex.Store. Options :args - A list of command line arguments :env - A map of environment variables :preopen - A list of Wasmex.Wasi.PreopenOptions to preopen directories :stdin - A Wasmex.Pipe to use as stdin :stdout - A Wasmex.Pipe to use as stdout :stderr - A Wasmex.Pipe to use as stderr Example iex&gt; { :ok , stdin } = Wasmex.Pipe . new ( ) iex&gt; Wasmex.Store . new_wasi ( % WasiOptions { ...&gt; args : [ &quot;first param&quot; , &quot;second param&quot; ] , ...&gt; env : %{ &quot;env_key&quot; =&gt; &quot;env_value&quot; } , ...&gt; preopen : [ ...&gt; % Wasmex.Wasi.PreopenOptions { path : &quot;/tmp&quot; } ...&gt; ] , ...&gt; stdin : stdin , ...&gt; } )","ref":"Wasmex.Wasi.WasiOptions.html","title":"Wasmex.Wasi.WasiOptions","type":"module"},{"doc":"","ref":"Wasmex.Wasi.WasiOptions.html#t:t/0","title":"Wasmex.Wasi.WasiOptions.t/0","type":"type"}]